@@ -41,7 +41,7 @@
 #include "egg/egg-error.h"
 #include "egg/egg-unix-credentials.h"
 
-#include "gck/gck.h"
+#include <gck/gck.h>
 
 #include "pkcs11/pkcs11i.h"
 
@@ -60,6 +60,7 @@ struct _GkdSecretService {
 	gchar *match_rule;
 	GkdSecretObjects *objects;
 	GHashTable *aliases;
+	GckSession *internal_session;
 };
 
 typedef struct _ServiceClient {
@@ -804,6 +805,52 @@ service_method_unlock_with_master_password (GkdSecretService *self, DBusMessage
 	return reply;
 }
 
+static void
+on_each_path_append_to_array (GkdSecretObjects *self,
+                              const gchar *path,
+                              GckObject *object,
+                              gpointer user_data)
+{
+	GPtrArray *array = user_data;
+	g_ptr_array_add (array, g_strdup (path));
+}
+
+static DBusMessage *
+service_introspect (GkdSecretService *self,
+                    DBusMessage *message)
+{
+	GPtrArray *names;
+	DBusMessage *reply;
+	ServiceClient *client;
+	const gchar *caller;
+	const gchar *path;
+	GHashTableIter iter;
+
+	names = g_ptr_array_new_with_free_func (g_free);
+	gkd_secret_objects_foreach_collection (self->objects, message,
+	                                       on_each_path_append_to_array,
+	                                       names);
+
+	/* Lookup all sessions and prompts for this client */
+	caller = dbus_message_get_sender (message);
+	if (caller != NULL) {
+		client = g_hash_table_lookup (self->clients, caller);
+		if (client != NULL) {
+			g_hash_table_iter_init (&iter, client->dispatch);
+			while (g_hash_table_iter_next (&iter, (gpointer *)&path, NULL))
+				g_ptr_array_add (names, g_strdup (path));
+		}
+	}
+
+	g_ptr_array_add (names, NULL);
+
+	reply = gkd_dbus_introspect_handle (message, gkd_secret_introspect_collection,
+	                                    (const gchar **)names->pdata);
+
+	g_ptr_array_unref (names);
+	return reply;
+}
+
 static DBusMessage*
 service_message_handler (GkdSecretService *self, DBusMessage *message)
 {
@@ -874,12 +921,31 @@ service_message_handler (GkdSecretService *self, DBusMessage *message)
 	else if (dbus_message_is_method_call (message, DBUS_INTERFACE_PROPERTIES, "GetAll"))
 		return service_property_getall (self, message);
 
+	/* org.freedesktop.DBus.Introspectable.Introspect() */
 	else if (dbus_message_has_interface (message, DBUS_INTERFACE_INTROSPECTABLE))
-		return gkd_dbus_introspect_handle (message, gkd_secret_introspect_service);
+		return service_introspect (self, message);
 
 	return NULL;
 }
 
+static gboolean
+root_dispatch_message (GkdSecretService *self,
+                       DBusMessage *message)
+{
+	DBusMessage *reply = NULL;
+
+	if (dbus_message_has_interface (message, DBUS_INTERFACE_INTROSPECTABLE))
+		reply = gkd_dbus_introspect_handle (message, gkd_secret_introspect_root, NULL);
+
+	if (reply != NULL) {
+		dbus_connection_send (self->connection, reply, NULL);
+		dbus_message_unref (reply);
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
 static void
 service_dispatch_message (GkdSecretService *self, DBusMessage *message)
 {
@@ -986,6 +1052,11 @@ gkd_secret_service_filter_handler (DBusConnection *conn, DBusMessage *message, g
 
 	/* Dispatch any method call on our interfaces, for our objects */
 	case DBUS_MESSAGE_TYPE_METHOD_CALL:
+		if (path != NULL && g_str_equal (path, "/")) {
+			if (root_dispatch_message (self, message))
+				return DBUS_HANDLER_RESULT_HANDLED;
+		}
+
 		if (object_path_has_prefix (path, SECRET_SERVICE_PATH)) {
 			interface = dbus_message_get_interface (message);
 			if (interface == NULL ||
@@ -1092,6 +1163,11 @@ gkd_secret_service_dispose (GObject *obj)
 		self->connection = NULL;
 	}
 
+	if (self->internal_session) {
+		dispose_and_unref (self->internal_session);
+		self->internal_session = NULL;
+	}
+
 	G_OBJECT_CLASS (gkd_secret_service_parent_class)->dispose (obj);
 }
 
@@ -1202,14 +1278,32 @@ gkd_secret_service_get_pkcs11_slot (GkdSecretService *self)
 	return gkd_secret_objects_get_pkcs11_slot (self->objects);
 }
 
+static gboolean
+log_into_pkcs11_session (GckSession *session, GError **error)
+{
+	GckTokenInfo *info;
+	GckSlot *slot;
+	gboolean login;
+
+	/* Perform the necessary 'user' login to secrets token. Doesn't unlock anything */
+	slot = gck_session_get_slot (session);
+	info = gck_slot_get_token_info (slot);
+	login = info && (info->flags & CKF_LOGIN_REQUIRED);
+	gck_token_info_free (info);
+	g_object_unref (slot);
+
+	if (login && !gck_session_login (session, CKU_USER, NULL, 0, NULL, error))
+		return FALSE;
+
+	return TRUE;
+}
+
 GckSession*
 gkd_secret_service_get_pkcs11_session (GkdSecretService *self, const gchar *caller)
 {
 	ServiceClient *client;
 	GError *error = NULL;
-	GckTokenInfo *info;
 	GckSlot *slot;
-	gboolean login;
 
 	g_return_val_if_fail (GKD_SECRET_IS_SERVICE (self), NULL);
 	g_return_val_if_fail (caller, NULL);
@@ -1230,11 +1324,7 @@ gkd_secret_service_get_pkcs11_session (GkdSecretService *self, const gchar *call
 			return NULL;
 		}
 
-		/* Perform the necessary 'user' login to secrets token. Doesn't unlock anything */
-		info = gck_slot_get_token_info (slot);
-		login = info && (info->flags & CKF_LOGIN_REQUIRED);
-		gck_token_info_free (info);
-		if (login && !gck_session_login (client->pkcs11_session, CKU_USER, NULL, 0, NULL, &error)) {
+		if (!log_into_pkcs11_session (client->pkcs11_session, &error)) {
 			g_warning ("couldn't log in to pkcs11 session for secret service: %s",
 			           egg_error_message (error));
 			g_clear_error (&error);
@@ -1247,6 +1337,39 @@ gkd_secret_service_get_pkcs11_session (GkdSecretService *self, const gchar *call
 	return client->pkcs11_session;
 }
 
+GckSession*
+gkd_secret_service_internal_pkcs11_session (GkdSecretService *self)
+{
+	GError *error = NULL;
+	GckSlot *slot;
+
+	g_return_val_if_fail (GKD_SECRET_IS_SERVICE (self), NULL);
+
+	if (self->internal_session)
+		return self->internal_session;
+
+	slot = gkd_secret_service_get_pkcs11_slot (self);
+	self->internal_session = gck_slot_open_session_full (slot, GCK_SESSION_READ_WRITE,
+	                                                     0, NULL, NULL, NULL, &error);
+	if (!self->internal_session) {
+		g_warning ("couldn't open pkcs11 session for secret service: %s",
+		           egg_error_message (error));
+		g_clear_error (&error);
+		return NULL;
+	}
+
+	if (!log_into_pkcs11_session (self->internal_session, &error)) {
+		g_warning ("couldn't log in to pkcs11 session for secret service: %s",
+		           egg_error_message (error));
+		g_clear_error (&error);
+		g_object_unref (self->internal_session);
+		self->internal_session = NULL;
+		return NULL;
+	}
+
+	return self->internal_session;
+}
+
 GkdSecretSession*
 gkd_secret_service_lookup_session (GkdSecretService *self, const gchar *path,
                                    const gchar *caller)
