@@ -119,20 +119,52 @@ aes_derive_key (GckSession *session, GckObject *priv_key,
                 gconstpointer input, gsize n_input, GckObject **aes_key)
 {
 	GError *error = NULL;
-	GckMechanism mech = { CKM_DH_PKCS_DERIVE, input, n_input };
+	GckMechanism mech;
 	GckAttributes *attrs;
+	GckObject *dh_key;
+
+	/*
+	 * First we have to generate a secret key from the DH key. The
+	 * length of this key depends on the size of our DH prime
+	 */
+
+	mech.type = CKM_DH_PKCS_DERIVE;
+	mech.parameter = input;
+	mech.n_parameter = n_input;
+
+	attrs = gck_attributes_new ();
+	gck_attributes_add_ulong (attrs, CKA_CLASS, CKO_SECRET_KEY);
+	gck_attributes_add_ulong (attrs, CKA_KEY_TYPE, CKK_GENERIC_SECRET);
+
+	dh_key = gck_session_derive_key_full (session, priv_key, &mech, attrs, NULL, &error);
+
+	gck_attributes_unref (attrs);
+
+	if (!dh_key) {
+		g_warning ("couldn't derive key from dh key pair: %s", egg_error_message (error));
+		g_clear_error (&error);
+		return FALSE;
+	}
+
+	/*
+	 * Now use HKDF to generate our AES key.
+	 */
+
+	mech.type = CKM_G_HKDF_SHA256_DERIVE;
+	mech.parameter = NULL;
+	mech.n_parameter = 0;
 
 	attrs = gck_attributes_new ();
 	gck_attributes_add_ulong (attrs, CKA_VALUE_LEN, 16UL);
 	gck_attributes_add_ulong (attrs, CKA_CLASS, CKO_SECRET_KEY);
 	gck_attributes_add_ulong (attrs, CKA_KEY_TYPE, CKK_AES);
 
-	*aes_key = gck_session_derive_key_full (session, priv_key, &mech, attrs, NULL, &error);
-
+	*aes_key = gck_session_derive_key_full (session, dh_key, &mech, attrs, NULL, &error);
 	gck_attributes_unref (attrs);
+	g_object_unref (dh_key);
 
 	if (!*aes_key) {
-		g_warning ("couldn't derive aes key from dh key pair: %s", egg_error_message (error));
+		g_warning ("couldn't derive aes key from dh key: %s", egg_error_message (error));
 		g_clear_error (&error);
 		return FALSE;
 	}
@@ -278,12 +310,13 @@ gkd_secret_session_real_dispatch_message (GkdSecretDispatch *base, DBusMessage *
 	if (!caller || !g_str_equal (caller, self->caller))
 		g_return_val_if_reached (NULL);
 
-	/* org.freedesktop.Secrets.Session.Close() */
+	/* org.freedesktop.Secret.Session.Close() */
 	else if (dbus_message_is_method_call (message, SECRET_SESSION_INTERFACE, "Close"))
 		return session_method_close (self, message);
 
+	/* org.freedesktop.DBus.Introspectable.Introspect() */
 	else if (dbus_message_has_interface (message, DBUS_INTERFACE_INTROSPECTABLE))
-		return gkd_dbus_introspect_handle (message, gkd_secret_introspect_session);
+		return gkd_dbus_introspect_handle (message, gkd_secret_introspect_session, NULL);
 
 	return NULL;
 }
@@ -529,7 +562,7 @@ gkd_secret_session_handle_open (GkdSecretSession *self, DBusMessage *message)
 			                               "The session algorithm input argument was invalid");
 		reply = plain_negotiate (self, message);
 
-	} else if (g_str_equal (algorithm, "dh-ietf1024-aes128-cbc-pkcs7")) {
+	} else if (g_str_equal (algorithm, "dh-ietf1024-sha256-aes128-cbc-pkcs7")) {
 		if (!g_str_equal ("ay", dbus_message_iter_get_signature (&variant)))
 			return dbus_message_new_error (message, DBUS_ERROR_INVALID_ARGS,
 			                               "The session algorithm input argument was invalid");
@@ -538,7 +571,7 @@ gkd_secret_session_handle_open (GkdSecretSession *self, DBusMessage *message)
 		reply = aes_negotiate (self, message, input, n_input);
 
 	} else {
-		reply = dbus_message_new_error_printf (message, SECRET_ERROR_NOT_SUPPORTED,
+		reply = dbus_message_new_error_printf (message, DBUS_ERROR_NOT_SUPPORTED,
 		                                       "The algorithm '%s' is not supported", algorithm);
 	}
 
