@@ -60,6 +60,7 @@
 
 #include <glib.h>
 #include <glib/gi18n.h>
+#include <glib-object.h>
 
 #include <gcrypt.h>
 
@@ -78,6 +79,8 @@ typedef int socklen_t;
 #define GKD_COMP_SSH        "ssh"
 #define GKD_COMP_GPG        "gpg"
 
+EGG_SECURE_DECLARE (daemon_main);
+
 /* -----------------------------------------------------------------------------
  * COMMAND LINE
  */
@@ -322,6 +325,58 @@ prepare_logging ()
 	g_set_printerr_handler (printerr_handler);
 }
 
+#ifdef WITH_DEBUG
+
+static void
+dump_diagnostics (void)
+{
+	egg_secure_rec *records;
+	egg_secure_rec *rec;
+	unsigned int count, i;
+	GHashTable *table;
+	GHashTableIter iter;
+	gsize request = 0;
+	gsize block = 0;
+
+	g_printerr ("------------------- Secure Memory --------------------\n");
+	g_printerr (" Tag                          Used            Space\n");
+	g_printerr ("------------------------------------------------------\n");
+
+	records = egg_secure_records (&count);
+	table = g_hash_table_new (g_str_hash, g_str_equal);
+	for (i = 0; i < count; i++) {
+		if (!records[i].tag)
+			records[i].tag = "<unused>";
+		rec = g_hash_table_lookup (table, records[i].tag);
+		if (rec == NULL)
+			g_hash_table_insert (table, (gchar *)records[i].tag, &records[i]);
+		else {
+			rec->block_length += records[i].block_length;
+			rec->request_length += records[i].request_length;
+		}
+		block += records[i].block_length;
+		request += records[i].request_length;
+	}
+
+	g_hash_table_iter_init (&iter, table);
+	while (g_hash_table_iter_next (&iter, NULL, (gpointer *)&rec))
+		g_printerr (" %-20s %12lu %16lu\n", rec->tag,
+		            (unsigned long)rec->request_length,
+		            (unsigned long)rec->block_length);
+
+	if (count > 0)
+		g_printerr ("------------------------------------------------------\n");
+
+	g_printerr (" %-20s %12lu %16lu\n", "Total",
+	            (unsigned long)request, (unsigned long)block);
+	g_printerr ("------------------------------------------------------\n");
+
+	g_hash_table_destroy (table);
+	free (records);
+}
+
+#endif /* WITH_DEBUG */
+
 /* -----------------------------------------------------------------------------
  * SIGNALS
  */
@@ -343,6 +398,11 @@ signal_thread (gpointer user_data)
 		}
 
 		switch (sig) {
+		case SIGUSR1:
+#ifdef WITH_DEBUG
+			dump_diagnostics ();
+#endif /* WITH_DEBUG */
+			break;
 		case SIGPIPE:
 			/* Ignore */
 			break;
@@ -378,6 +438,7 @@ setup_signal_handling (GMainLoop *loop)
 	sigaddset (&signal_set, SIGPIPE);
 	sigaddset (&signal_set, SIGHUP);
 	sigaddset (&signal_set, SIGTERM);
+	sigaddset (&signal_set, SIGUSR1);
 	pthread_sigmask (SIG_BLOCK, &signal_set, NULL);
 
 	res = pthread_create (&sig_thread, NULL, signal_thread, loop);
@@ -814,7 +875,7 @@ main (int argc, char *argv[])
 	 */
 	gkd_capability_obtain_capability_and_drop_privileges ();
 
-#ifdef WITH_TESTABLE
+#ifdef WITH_STRICT
 	g_setenv ("DBUS_FATAL_WARNINGS", "1", FALSE);
 	if (!g_getenv ("G_DEBUG"))
 		g_log_set_always_fatal (G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING);
@@ -845,7 +906,7 @@ main (int argc, char *argv[])
 	if (run_version) {
 		g_print ("gnome-keyring-daemon: %s\n", VERSION);
 		g_print ("testing: %s\n",
-#ifdef WITH_TESTABLE
+#ifdef WITH_DEBUG
 		         "enabled");
 #else
 		         "disabled");
