@@ -28,7 +28,7 @@
 #include "egg/egg-error.h"
 #include "egg/egg-secure-memory.h"
 
-#include "gck/gck.h"
+#include <gck/gck.h>
 
 #include <glib.h>
 #include <glib-object.h>
@@ -88,7 +88,7 @@ main(int argc, char *argv[])
 		return 1;
 	}
 
-	module = gck_module_initialize (argv[1], &error);
+	module = gck_module_initialize (argv[1], NULL, &error);
 	if (!module) {
 		g_message ("couldn't load pkcs11 module: %s", egg_error_message (error));
 		g_clear_error (&error);
