@@ -26,7 +26,7 @@
 
 #include "egg/egg-secure-memory.h"
 
-#include "gcr/gcr-unlock-options.h"
+#include <gcr/gcr-unlock-options.h>
 
 #include "gkm/gkm-attributes.h"
 #include "gkm/gkm-util.h"
@@ -537,7 +537,7 @@ get_unlock_options_from_prompt (GkmWrapPrompt *self, CK_ULONG_PTR n_options)
 static void
 set_unlock_options_on_prompt (GkmWrapPrompt *self, CK_ATTRIBUTE_PTR options, CK_ULONG n_options)
 {
-	const gchar *choice;
+	const gchar *choice = 0;
 	gboolean have_ttl = FALSE;
 	gboolean bval;
 	gulong uval;
